"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/FlowchartGraphicsView.py
"""
from pyqtgraph.Qt import QtGui, QtCore
from pyqtgraph.widgets.GraphicsView import GraphicsView
from pyqtgraph.graphicsItems.ViewBox import ViewBox


class RxView(GraphicsView):

    sigHoverOver = QtCore.Signal(object)
    sigClicked = QtCore.Signal(object)

    def __init__(self, widget, *args):
        GraphicsView.__init__(self, *args, useOpenGL=False)
        self.setBackground((255, 255, 255))
        self._vb = RxViewBox(widget, lockAspect=True, invertY=True)
        self.setCentralItem(self._vb)
        self.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)

    def viewBox(self):
        return self._vb


class RxViewBox(ViewBox):
    def __init__(self, widget, *args, **kwargs):
        ViewBox.__init__(self, *args, **kwargs)
        self.widget = widget

    def getMenu(self, ev):
        # called by ViewBox to create a new context menu
        self._fc_menu = QtGui.QMenu()
        self._subMenus = self.getContextMenus(ev)
        for menu in self._subMenus:
            self._fc_menu.addMenu(menu)
        return self._fc_menu

    def getContextMenus(self, ev):
        # called by scene to add menus on to someone else's context menu
        menus = self.widget.buildMenu(ev.scenePos())
        menus.append(ViewBox.getMenu(self, ev))
        return menus



import yaml
import inspect

from functools import wraps, partial
from copy import deepcopy

from pyqtgraph import ComboBox, SpinBox
from pyqtgraph.Qt import QtWidgets

from eagerx_gui import configuration
from eagerx_gui.utils import tryeval
from eagerx.utils.utils import (
    get_attribute_from_module,
    get_module_type_string,
    get_opposite_msg_cls,
)
from eagerx.core.register import REVERSE_REGISTRY


def exception_handler(function_to_decorate):
    wraps(function_to_decorate)

    def exception_handler_wrapper(*args, graph_backup=None, dialog_title="Invalid", **kwargs):
        assert graph_backup is not None, "Graph should be defined for recovering state."
        state_copy = deepcopy(graph_backup._state)
        try:
            return function_to_decorate(*args, **kwargs)
        except Exception as e:
            graph_backup._state = state_copy
            graph_backup.load_state(clear=True)
            error_window = QtWidgets.QDialog(graph_backup.widget().cwWin)
            error_window.setWindowTitle(dialog_title)
            layout = QtWidgets.QGridLayout()
            label = QtWidgets.QLabel(str(e))
            layout.addWidget(label)
            error_window.setLayout(layout)
            error_window.exec()
            raise

    return exception_handler_wrapper


class NodeCreationDialog(QtWidgets.QDialog):
    def __init__(self, name, node_type, parent):
        super().__init__(parent)
        self.setWindowTitle("Create {}".format(name))
        self.mapping = {}
        self.node_type = node_type

        signature = node_type["entity_cls"].get_spec(node_type["entity_id"], verbose=False)
        parameters = signature.parameters

        self.layout = QtWidgets.QGridLayout()
        self.labels = []
        self.widgets = []
        row = 0

        required_args = {}
        optional_args = {}

        for key in parameters.keys():
            if parameters[key].default is inspect._empty:
                required_args[key] = parameters[key]
            else:
                optional_args[key] = parameters[key]

        required_args_label = QtWidgets.QLabel("Required Arguments")
        self.layout.addWidget(required_args_label, row, 0)
        self.labels.append(required_args_label)
        row += 1

        for key, parameter in required_args.items():
            if key == "name":
                value = name
            elif key == "rate":
                value = 1.0
            else:
                value = None
            self.add_widget(key, value, parameter, row)
            row += 1

        optional_args_label = QtWidgets.QLabel("Optional Arguments")
        self.layout.addWidget(optional_args_label, row, 0)
        self.labels.append(optional_args_label)
        row += 1

        for key, parameter in optional_args.items():
            value = parameter.default
            self.add_widget(key, value, parameter, row)
            row += 1

        self.setLayout(self.layout)

    def open(self):
        self.exec()
        return self.mapping

    def add_widget(self, key, value, parameter, row):
        label = QtWidgets.QLabel(str(parameter).split("=")[0].strip())
        if parameter.annotation is bool:
            value = value if value is not None else True
            items = ["True", "False"]
            widget = ComboBox(items=items, default=str(value))
            widget.activated.connect(partial(self.combo_box_value_changed, key=key, items=items))
        elif parameter.annotation is int:
            widget = SpinBox(value=value, int=True, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif parameter.annotation is float:
            value = value if value is not None else 0.0
            widget = SpinBox(value=value, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif parameter.annotation is str:
            widget = QtWidgets.QLineEdit(str(value))
            widget.textChanged.connect(partial(self.text_changed, key=key))
        elif isinstance(value, bool):
            items = ["True", "False"]
            widget = ComboBox(items=items, default=str(value))
            widget.activated.connect(partial(self.combo_box_value_changed, key=key, items=items))
        elif isinstance(value, int):
            widget = SpinBox(value=value, int=True, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif isinstance(value, float):
            widget = SpinBox(value=value, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        else:
            widget = QtWidgets.QLineEdit(str(value))
            widget.textChanged.connect(partial(self.text_changed, key=key))
        self.mapping[key] = value
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        self.layout.addWidget(label, row, 0)
        self.layout.addWidget(widget, row, 1)
        self.labels.append(label)
        self.widgets.append(widget)

    def combo_box_value_changed(self, int, items, key):
        value = items[int]
        self.mapping[key] = value

    def value_changed(self, widget, key):
        self.mapping[key] = widget.value()

    def text_changed(self, text, key):
        self.mapping[key] = tryeval(text)


class ParamWindow(QtWidgets.QDialog):
    def __init__(self, node, term=None):
        self.parent = node.graph.widget().cwWin
        super().__init__(self.parent)
        self.node = node
        self.node_type = node.node_type
        self.library = node.graph.library
        self.is_term = term is not None
        self.entity = term if self.is_term else node
        self.entity_type = "term" if self.is_term else "node"

        name = term.terminal_name if self.is_term else node.name
        self.setWindowTitle("Parameters {}".format(name))

        self.layout = QtWidgets.QGridLayout()
        self.labels = []
        self.widgets = []
        self.params_changed = {}
        row = 0

        self.gui_widgets = configuration.ENGINE_GUI_WIDGETS if self.node_type == "engine_node" else configuration.GUI_WIDGETS
        for key, value in self.entity.params().items():
            if key in self.gui_widgets[self.entity_type]["hide"]["all"]:
                continue
            elif (
                self.node_type in self.gui_widgets[self.entity_type]["hide"]
                and key in self.gui_widgets[self.entity_type]["hide"][self.node_type]
            ):
                continue
            elif (
                self.is_term
                and self.entity.terminal_type in self.gui_widgets[self.entity_type]["hide"]
                and key in self.gui_widgets[self.entity_type]["hide"][self.entity.terminal_type]
            ):
                continue
            elif (
                self.node_type == "engine_node"
                and self.is_term
                and self.entity.node.name == "sensors"
                and key == "external_rate"
            ):
                continue
            else:
                self.add_widget(key, value, row)
                row += 1
        if row == 0:
            label = QtWidgets.QLabel("No parameters to show.")
            self.layout.addWidget(label, row, 0)
        self.setLayout(self.layout)

    def open(self):
        self.exec()
        valid = False
        while not valid:
            try:
                self.set_params(graph_backup=self.node.graph)
                self.node.graph.load_state(clear=True)
                valid = True
            except Exception:
                self.exec()

    @exception_handler
    def set_params(self):
        for key, value in self.params_changed.items():
            value = yaml.safe_load(str(value))
            self.entity.set_param(key, value)

    def add_widget(self, key, value, row):
        label = QtWidgets.QLabel(key)
        if self.is_term and key in "converter":
            button_string = value["converter_type"].split("/")[-1] if "converter_type" in value else "converter"
            widget = QtWidgets.QPushButton("Edit {}".format(button_string))
            widget.pressed.connect(partial(self.open_converter_dialog, button=widget))
        elif self.is_term and key == "space_converter":
            if self.entity.is_state:
                button_string = value["converter_type"].split("/")[-1] if "converter_type" in value else "space_converter"
                widget = QtWidgets.QPushButton("Edit {}".format(button_string))
                widget.pressed.connect(
                    partial(
                        self.open_converter_dialog,
                        button=widget,
                        is_space_converter=True,
                    )
                )
            else:
                widget = QtWidgets.QLineEdit(str(value))
                widget.setEnabled(False)
        elif key in self.gui_widgets[self.entity_type]["items"]:
            items = self.gui_widgets[self.entity_type]["items"][key]
            if isinstance(items, dict):
                item_converter = items
                index, items = list(items.values()).index(value), list(items.keys())
                value = items[index]
            else:
                item_converter = None
            widget = ComboBox(items=items, default=str(value))
            widget.activated.connect(
                partial(
                    self.combo_box_value_changed,
                    key=key,
                    items=items,
                    item_converter=item_converter,
                )
            )
        elif isinstance(value, bool):
            items = ["True", "False"]
            widget = ComboBox(items=items, default=str(value))
            widget.activated.connect(partial(self.combo_box_value_changed, key=key, items=items))
        elif isinstance(value, int):
            widget = SpinBox(value=value, int=True, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif isinstance(value, float):
            widget = SpinBox(value=value, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        else:
            widget = QtWidgets.QLineEdit(str(value))
            widget.textChanged.connect(partial(self.text_changed, key=key))
        if key in self.gui_widgets[self.entity_type]["constant"]["all"] or (
            self.node_type in self.gui_widgets[self.entity_type]["constant"]
            and key in self.gui_widgets[self.entity_type]["constant"][self.node_type]
        ):
            widget.setEnabled(False)
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        self.layout.addWidget(label, row, 0)
        self.layout.addWidget(widget, row, 1)
        self.labels.append(label)

        # Nodes in in the EngineGraph Gui cannot be modified
        if self.node_type == "engine_node" and self.entity.name:
            widget.setEnabled(False)
        self.widgets.append(widget)

    def combo_box_value_changed(self, int, items, key, item_converter=None):
        value = items[int]
        if item_converter is not None:
            value = item_converter[value]
        self.params_changed[key] = value

    def value_changed(self, widget, key):
        self.params_changed[key] = widget.value()

    def text_changed(self, text, key):
        self.params_changed[key] = tryeval(text)

    def open_converter_dialog(self, button, is_space_converter=False):
        key = "space_converter" if is_space_converter else "converter"
        library = self.node.graph.library
        if "converter" in self.entity.params() and self.entity.params()["converter"] is not None:
            msg_type = get_opposite_msg_cls(self.entity.params()["msg_type"], self.entity.params()["converter"])
        else:
            msg_type = get_attribute_from_module(self.entity.params()["msg_type"])
        msg_type_in, msg_type_out = (msg_type, None) if self.entity.is_input else (None, msg_type)
        is_space_converter = is_space_converter or self.node.node_type in [
            "actions",
            "observations",
        ]
        converter_dialog = ConverterDialog(
            converter=self.entity.params()[key],
            parent=self.parent,
            library=library,
            msg_type_in=msg_type_in,
            msg_type_out=msg_type_out,
            is_space_converter=is_space_converter,
        )
        converter = converter_dialog.open()
        converter_dialog.close()
        self.params_changed[key] = converter
        button_string = converter["converter_type"].split("/")[-1] if "converter_type" in converter else key
        button.setText("Edit {}".format(button_string))


class ConverterDialog(QtWidgets.QDialog):
    def __init__(
        self,
        converter,
        parent,
        library,
        msg_type_in,
        msg_type_out=None,
        is_space_converter=False,
    ):
        super().__init__(parent)
        self.parent = parent
        self.msg_type_in = msg_type_in
        self.msg_type_out = msg_type_out
        self.is_space_converter = is_space_converter
        self.converter = converter
        self.library = library
        self.setWindowTitle("Converter Parameters")
        self.layout = QtWidgets.QGridLayout()
        self.labels = []
        self.widgets = []

        converter_class = get_attribute_from_module(self.converter["converter_type"])
        converter_id = REVERSE_REGISTRY[converter_class.spec]
        (
            converter_id,
            available_converters,
            required_args,
            optional_args,
        ) = self.get_parameters(converter_id)

        self.add_widget(
            key="Converter Class",
            value=converter_id,
            parameter=None,
            row=0,
            items=available_converters,
        )
        self.add_argument_widgets(required_args, optional_args)
        self.setLayout(self.layout)

    def open(self):
        self.exec()
        valid = False
        while not valid:
            try:
                for key, value in self.converter.items():
                    self.converter[key] = yaml.safe_load(str(value))
                get_attribute_from_module(self.converter["converter_type"])
                valid = True
            except Exception as e:
                error_window = QtWidgets.QDialog(self.parent)
                error_window.setWindowTitle("Invalid Converter")
                layout = QtWidgets.QGridLayout()
                label = QtWidgets.QLabel(str(e))
                layout.addWidget(label)
                error_window.setLayout(layout)
                error_window.exec()
                self.exec()
        return self.converter

    def add_argument_widgets(self, required_args, optional_args):
        row = 1
        required_args_label = QtWidgets.QLabel("Required Converter Arguments")
        self.layout.addWidget(required_args_label, row, 0)
        self.labels.append(required_args_label)
        row += 1

        for key, parameter in required_args.items():
            if key in self.converter:
                value = self.converter[key]
            else:
                value = None
            self.add_widget(key, value, parameter, row)
            row += 1

        optional_args_label = QtWidgets.QLabel("Optional Converter Arguments")
        self.layout.addWidget(optional_args_label, row, 0)
        self.labels.append(optional_args_label)
        row += 1

        for key, parameter in optional_args.items():
            if key in self.converter:
                value = self.converter[key]
            else:
                value = parameter.default
            self.add_widget(key, value, parameter, row)
            row += 1

    def get_parameters(self, converter_id):
        converter = deepcopy(self.converter)

        available_converters = {}
        required_args = {}
        optional_args = {}

        if self.is_space_converter:
            cnvrtr_types = ["SpaceConverter"]
        elif None in [self.msg_type_in, self.msg_type_out]:
            cnvrtr_types = ["BaseConverter", "Processor", "Converter"]
        elif self.msg_type_in == self.msg_type_out:
            cnvrtr_types = ["BaseConverter", "Processor"]
        else:
            cnvrtr_types = ["Converter"]

        for cnvrtr_type in cnvrtr_types:
            if cnvrtr_type not in self.library:
                continue
            for cnvrtr in self.library[cnvrtr_type]:
                cnvrtr_cls = cnvrtr["cls"]
                if cnvrtr_type == "Processor":
                    if not cnvrtr_cls.MSG_TYPE in [self.msg_type_in, self.msg_type_out]:
                        continue
                elif cnvrtr_type in ["Converter", "SpaceConverter"]:
                    if self.msg_type_in is not None and not self.msg_type_in in [
                        cnvrtr_cls.MSG_TYPE_A,
                        cnvrtr_cls.MSG_TYPE_B,
                    ]:
                        continue
                    elif self.msg_type_out is not None and not self.msg_type_out in [
                        cnvrtr_cls.MSG_TYPE_A,
                        cnvrtr_cls.MSG_TYPE_B,
                    ]:
                        continue
                available_converters[cnvrtr["entity_id"]] = {
                    "spec": cnvrtr["entity_cls"].get_spec(cnvrtr["entity_id"], verbose=False),
                    "cls": cnvrtr_cls,
                }

        available_converters_list = list(available_converters.keys())

        if converter_id not in available_converters.keys():
            if "Identity" in available_converters:
                converter_id = "Identity"
            elif len(available_converters_list) > 0:
                converter_id = available_converters_list[0]
            else:
                converter_id = None

        if converter_id is not None:
            parameters = available_converters[converter_id]["spec"].parameters
            for key in parameters.keys():
                if parameters[key].default is inspect._empty:
                    required_args[key] = parameters[key]
                else:
                    optional_args[key] = parameters[key]
            invalid_arguments = []
            for key in converter.keys():
                if key not in parameters.keys():
                    invalid_arguments.append(key)
            for key in invalid_arguments:
                converter.pop(key)

        if converter_id is not None:
            converter["converter_type"] = get_module_type_string(available_converters[converter_id]["cls"])
        else:
            converter = None
        self.converter = converter
        return converter_id, available_converters_list, required_args, optional_args

    def add_widget(self, key, value, parameter, row, items=None):
        if parameter is not None:
            label = QtWidgets.QLabel(str(parameter).split("=")[0].strip())
        else:
            label = QtWidgets.QLabel(str(key))
        if parameter is None:
            widget = ComboBox(items=items, default=value)
            widget.activated.connect(partial(self.class_changed, items=items))
        elif parameter.annotation is bool:
            value = value if value is not None else True
            items = ["True", "False"]
            self.converter[key] = value
            widget = ComboBox(items=items, default=str(value))
            widget.activated.connect(partial(self.combo_box_value_changed, key=key, items=items))
        elif parameter.annotation is int:
            value = value if value is not None else 0
            self.converter[key] = value
            widget = SpinBox(value=value, int=True, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif parameter.annotation is float:
            value = value if value is not None else 0.0
            self.converter[key] = value
            widget = SpinBox(value=value, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif parameter.annotation is str:
            widget = QtWidgets.QLineEdit(str(value))
            widget.textChanged.connect(partial(self.text_changed, key=key))
        else:
            widget = QtWidgets.QLineEdit(str(value))
            widget.textChanged.connect(partial(self.argument_changed, key=key))

        self.layout.addWidget(label, row, 0)
        self.layout.addWidget(widget, row, 1)
        self.labels.append(label)
        self.widgets.append(widget)

    def class_changed(self, int, items):
        converter_id = items[int]

        for label in self.labels[1:]:
            self.layout.removeWidget(label)
            label.close()

        for widget in self.widgets[1:]:
            self.layout.removeWidget(widget)
            widget.close()

        self.labels = self.labels[:2]
        self.widgets = self.widgets[:2]

        _, _, required_args, optional_args = self.get_parameters(converter_id)

        self.add_argument_widgets(required_args, optional_args)

    def argument_changed(self, text, key):
        self.converter[key] = text

    def text_changed(self, text, key):
        self.params[key] = tryeval(text)


class ConnectionDialog(QtWidgets.QDialog):
    def __init__(self, input_term, output_term, **kwargs):
        self.parent = input_term.node.graph.widget().cwWin
        super().__init__(self.parent)
        self.setWindowTitle("Connection Parameters")
        self.library = input_term.node.graph.library
        self.input_term = input_term
        self.output_term = output_term
        self.layout = QtWidgets.QGridLayout()
        self.params = {}
        self.labels = []
        self.widgets = []
        row = 0
        params = deepcopy(input_term.params())
        for key, value in kwargs.items():
            if value is not None:
                params[key] = value
        for key, value in params.items():
            if key in inspect.getfullargspec(input_term.node.graph.connect).args:
                self.params[key] = value
                self.add_widget(key, value, row)
                row += 1
        self.setLayout(self.layout)

    def open(self):
        self.exec()
        for widget in self.widgets:
            widget.setEnabled(False)
        return self.params

    def add_widget(self, key, value, row):
        label = QtWidgets.QLabel(key)
        if key == "converter":
            button_string = value["converter_type"].split("/")[-1] if "converter_type" in value else "converter"
            widget = QtWidgets.QPushButton("Edit {}".format(button_string))
            widget.pressed.connect(partial(self.open_converter_dialog, button=widget))
        elif isinstance(value, bool):
            items = ["True", "False"]
            widget = ComboBox(items=items, default=str(value))
            widget.activated.connect(partial(self.combo_box_value_changed, key=key, items=items))
        elif isinstance(value, int):
            widget = SpinBox(value=value, int=True, dec=True)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        elif isinstance(value, float):
            widget = SpinBox(value=value)
            widget.sigValueChanged.connect(partial(self.value_changed, key=key))
        else:
            widget = QtWidgets.QLineEdit(str(value))
            widget.textChanged.connect(partial(self.text_changed, key=key))
        for grid_object in [label, widget]:
            font = grid_object.font()
            font.setPointSize(12)
            grid_object.setFont(font)
        self.layout.addWidget(label, row, 0)
        self.layout.addWidget(widget, row, 1)
        self.labels.append(label)
        self.widgets.append(widget)

    def open_converter_dialog(self, button):
        is_space_converter = self.input_term.node.node_type == "observations"
        library = self.input_term.node.graph.library
        if self.input_term.node.node_type == "observations":
            msg_type_in = None
            if "converter" in self.output_term.params() and self.output_term.params()["converter"] is not None:
                msg_type_out = get_opposite_msg_cls(
                    self.output_term.params()["msg_type"],
                    self.output_term.params()["converter"],
                )
            else:
                msg_type_out = get_attribute_from_module(self.input_term.params()["msg_type"])
        elif self.output_term.node.node_type == "actions" and "msg_type" not in self.output_term.params():
            if "converter" in self.input_term.params() and self.input_term.params()["converter"] is not None:
                msg_type_in = get_opposite_msg_cls(
                    self.input_term.params()["msg_type"],
                    self.input_term.params()["converter"],
                )
            else:
                msg_type_in = get_attribute_from_module(self.input_term.params()["msg_type"])
            msg_type_out = msg_type_in
        else:
            if "converter" in self.input_term.params() and self.input_term.params()["converter"] is not None:
                msg_type_in = get_opposite_msg_cls(
                    self.input_term.params()["msg_type"],
                    self.input_term.params()["converter"],
                )
            else:
                msg_type_in = get_attribute_from_module(self.input_term.params()["msg_type"])
            if "converter" in self.output_term.params() and self.output_term.params()["converter"] is not None:
                msg_type_out = get_opposite_msg_cls(
                    self.output_term.params()["msg_type"],
                    self.output_term.params()["converter"],
                )
            else:
                msg_type_out = get_attribute_from_module(self.input_term.params()["msg_type"])
        converter_dialog = ConverterDialog(
            converter=self.params["converter"],
            parent=self.parent,
            library=library,
            msg_type_in=msg_type_in,
            msg_type_out=msg_type_out,
            is_space_converter=is_space_converter,
        )
        converter = converter_dialog.open()
        self.params["converter"] = converter
        converter_dialog.close()
        button_string = converter["converter_type"].split("/")[-1] if "converter_type" in converter else "converter"
        button.setText("Edit {}".format(button_string))

    def combo_box_value_changed(self, int, items, key):
        self.params[key] = items[int]

    def value_changed(self, widget, key):
        self.params[key] = widget.value()

    def text_changed(self, text, key):
        self.params[key] = tryeval(text)



"""
 Mostly copy paste from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/FlowchartCtrlTemplate_pyqt6.py
"""
from PyQt6 import QtCore, QtGui, QtWidgets
from pyqtgraph.widgets.FeedbackButton import FeedbackButton
from pyqtgraph.widgets.TreeWidget import TreeWidget


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(217, 499)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setVerticalSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.checkValidityBtn = FeedbackButton(Form)
        self.checkValidityBtn.setObjectName("checkValidityBtn")
        self.gridLayout.addWidget(self.checkValidityBtn, 4, 0, 1, 1)
        self.showChartBtn = QtWidgets.QPushButton(Form)
        self.showChartBtn.setCheckable(True)
        self.showChartBtn.setObjectName("showChartBtn")
        self.gridLayout.addWidget(self.showChartBtn, 4, 3, 1, 1)
        self.ctrlList = TreeWidget(Form)
        self.ctrlList.setObjectName("ctrlList")
        self.ctrlList.headerItem().setText(0, "1")
        self.ctrlList.header().setVisible(False)
        self.ctrlList.header().setStretchLastSection(False)
        self.gridLayout.addWidget(self.ctrlList, 3, 0, 1, 4)
        self.fileNameLabel = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setBold(True)
        self.fileNameLabel.setFont(font)
        self.fileNameLabel.setText("")
        self.fileNameLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.fileNameLabel.setObjectName("fileNameLabel")
        self.gridLayout.addWidget(self.fileNameLabel, 0, 1, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "EAGERx"))
        self.checkValidityBtn.setText(_translate("Form", "Check Validity"))
        self.showChartBtn.setText(_translate("Form", "Show Graph"))



from eagerx.utils.utils import get_attribute_from_module
import networkx as nx
import ast


def tryeval(val):
    try:
        val = ast.literal_eval(val)
    except Exception as e:
        if isinstance(e, ValueError):
            pass
        elif isinstance(e, SyntaxError):
            pass
        else:
            raise
    return val


def empty_gui_state():
    return dict(pos=None, linestyle=dict())


def get_yaml_type(yaml):
    if "node_type" in yaml:
        if "targets" in yaml:
            type = "reset_node"
        else:
            type = "node"
    else:
        type = "object"
    return type


def get_nodes_and_objects_library():
    from eagerx.core.register import REGISTRY

    library = dict()
    for entity_cls, entities in REGISTRY.items():
        library[entity_cls.__name__] = []
        for id, entry in entities.items():
            spec = entry["spec"]
            cls = get_attribute_from_module(entry["cls"])
            library[entity_cls.__name__].append({"entity_id": id, "spec": spec, "entity_cls": entity_cls, "cls": cls})
    return library


def get_connected_nodes(state, node):
    # Get a set of all the nodes that are connected directly to some node
    connected_nodes = set()
    for source, target in state["connects"]:
        if source[0] == node:
            connected_nodes.add(target[0])
        elif target[0] == node:
            connected_nodes.add(source[0])
    return connected_nodes


def is_connected_to_fixed(state, node, fixed_positions, nodes_to_ignore=None):
    # Check if node is connected to any fixed node (direct or indirect)
    if nodes_to_ignore is None:
        nodes_to_ignore = set()
    nodes_to_ignore.add(node)
    connected_nodes = get_connected_nodes(state, node)
    for connected_node in connected_nodes:
        if connected_node in nodes_to_ignore:
            continue
        elif connected_node in fixed_positions.keys():
            return True
        elif is_connected_to_fixed(state, connected_node, fixed_positions, nodes_to_ignore):
            return True
    return False


def add_pos_to_state(state, fixed_nodes=["env/actions", "env/observations", "env/render"]):
    # Position nodes using Fruchterman-Reingold force-directed algorithm.
    node_size = 150
    G = nx.Graph()
    fixed_positions = {}
    disconnected_y = node_size
    for node, params in state["nodes"].items():
        if node not in state["gui_state"]:
            state["gui_state"][node] = empty_gui_state()  # Add gui states here.
        gui_state = state["gui_state"][node]
        if gui_state.get("pos", None) is not None and len(gui_state["pos"]) == 2:
            fixed_positions[node] = gui_state["pos"]
        elif params["config"]["name"] in fixed_nodes:
            x_pos = node_size * (2 + max((len(state["nodes"].items()) - 3) // 2, 2))
            if params["config"]["name"] == fixed_nodes[0]:
                pos = [0, 0]
            elif params["config"]["name"] == fixed_nodes[1]:
                pos = [x_pos, 0]
            else:
                pos = [x_pos, node_size]
            fixed_positions[node] = pos
        elif not is_connected_to_fixed(state, node, fixed_positions):
            fixed_positions[node] = [0, disconnected_y]
            disconnected_y += node_size
        G.add_node(node)
    for source, target in state["connects"]:
        source_name, _, _ = source
        target_name, _, _ = target
        G.add_edge(source_name, target_name)
    position_dict = nx.spring_layout(G, k=node_size, pos=fixed_positions, fixed=fixed_positions.keys())
    for node, pos in position_dict.items():
        state["gui_state"][node]["pos"] = pos.tolist()
    return state


def add_pos_to_engine_state(state, fixed_nodes=["sensors", "actuators"]):
    # Position nodes using Fruchterman-Reingold force-directed algorithm.
    node_size = 150
    G = nx.Graph()
    fixed_positions = {}
    disconnected_y = node_size

    # First get two clusters of nodes, i.e. actions and observations cluster
    clusters = {}
    y_pos = {}
    for fixed_node in fixed_nodes:
        clusters[fixed_node] = get_connected_nodes(state, fixed_node)

    for key, cluster in clusters.items():
        y_pos[key] = -((len(cluster) - 1) / 2) * node_size

    for node, params in state["nodes"].items():
        if node not in state["gui_state"]:
            state["gui_state"][node] = empty_gui_state()  # Add gui states here.
        gui_state = state["gui_state"][node]
        if gui_state.get("pos", None) is not None and len(gui_state["pos"]) == 2:
            fixed_positions[node] = gui_state["pos"]
        elif params["config"]["name"] in fixed_nodes:
            if params["config"]["name"] == fixed_nodes[0]:
                pos = [0, 0]
            else:
                pos = [node_size, 0]
            fixed_positions[node] = pos
        else:
            for key, cluster in clusters.items():
                if node in cluster:
                    if key == fixed_nodes[0]:
                        x_pos = -node_size
                    else:
                        x_pos = 2 * node_size
                    fixed_positions[node] = [x_pos, y_pos[key]]
                    y_pos[key] += node_size
                    break
        G.add_node(node)
    for source, target in state["connects"]:
        source_name, _, _ = source
        target_name, _, _ = target
        G.add_edge(source_name, target_name)
    position_dict = nx.spring_layout(G, k=node_size, pos=fixed_positions, fixed=fixed_positions.keys())
    for node, pos in position_dict.items():
        state["gui_state"][node]["pos"] = pos.tolist()
    return state



"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Terminal.py
"""

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph import functions as fn
from pyqtgraph.Point import Point

from eagerx_gui import configuration
from eagerx_gui.pyqtgraph_utils import (
    exception_handler,
    ConnectionDialog,
    ParamWindow,
    ConverterDialog,
)
from eagerx.utils.utils import (
    get_module_type_string,
    get_opposite_msg_cls,
    get_attribute_from_module,
)
from eagerx.core.converters import Identity


class GuiTerminal(object):
    def __init__(self, node, name, pos=None):
        self.node = node
        self.name = name
        self.pos = pos
        self.optional = False
        self.connections = {}
        self.node_type = node.node_type

        name_split = name.split("/")
        self.terminal_type = name_split[0]
        self.terminal_name = name_split[-1]

        assert self.terminal_type in set.union(configuration.TERMS_IN, configuration.TERMS_OUT), (
            f"Invalid terminal type: {self.terminal_type}, "
            f"should be one of {set.union(configuration.TERMS_IN, configuration.TERMS_OUT)}"
        )

        self.is_input = self.terminal_type in configuration.TERMS_IN
        self.is_state = self.terminal_type in ["targets", "states"]
        self.is_feedthrough = self.terminal_type == "feedthroughs"
        self.is_removable = node.is_object or self.node_type in [
            "actions",
            "observations",
        ]
        self.is_renamable = self.node_type in ["actions", "observations"]
        self.is_connectable = not (self.node_type == "engine_node" or (self.terminal_type == "states" and not node.is_object))

        self._graphicsItem = TerminalGraphicsItem(self, parent=self.node.graphics_item())
        self.recolor()

    def connection_view(self):
        return self.node.graph.tuple_to_view(self.connection_tuple())

    def connection_tuple(self):
        return (self.node.name, self.terminal_type, self.terminal_name)

    def connected(self, term):
        """Called whenever this terminal has been connected to another.
        (note--this function is called on both terminals)"""
        self.node.connected(self, term)

    def disconnected(self, term):
        """Called whenever this terminal has been disconnected from another.
        (note--this function is called on both terminals)"""
        self.node.disconnected(self, term)

    def params(self):
        return self._get_params(graph_backup=self.node.graph)

    @exception_handler
    def _get_params(self):
        view = self.connection_view()
        params = view.to_dict()
        if self.terminal_type == "feedthroughs":
            view = self.node.graph.tuple_to_view((self.node.name, "outputs", self.terminal_name))
            output_params = view.to_dict()
            params["msg_type"] = output_params["msg_type"]
            params["space_converter"] = output_params["space_converter"]
        return params

    def set_param(self, parameter, value):
        self._set_param(parameter, value, graph_backup=self.node.graph)

    @exception_handler
    def _set_param(self, parameter, value):
        self.node.graph.set(value, self.connection_view(), parameter=parameter)

    def is_connected(self):
        return len(self.connections) > 0

    def graphics_item(self):
        return self._graphicsItem

    def connected_to(self, term):
        return term in self.connections

    def has_input(self):
        for t in self.connections:
            if not t.is_input:
                return True
        return False

    def input_terminals(self):
        """Return the terminal(s) that give input to this one."""
        return [t for t in self.connections if not t.is_input]

    def dependent_nodes(self):
        """Return the list of nodes which receive input from this terminal."""
        return set([t.node for t in self.connections if t.is_input])

    @exception_handler
    def connect_to(self, term, connection_item):
        assert not self.connected_to(term), "Already connected"
        assert term is not self, "Not connecting terminal to self"
        assert term.node is not self.node, "Cannot connect to terminal on same node."
        assert term.is_state == self.is_state, "Cannot connect states to other types than targets."
        assert not term.is_input == self.is_input, "Cannot connect input with input or output with output."
        for t in [self, term]:
            assert not (
                t.is_input and t.is_connected()
            ), "Cannot connect %s <-> %s: Terminal %s is already connected to %s" % (
                self,
                term,
                t,
                list(t.connections.keys()),
            )

        input_term, output_term = (self, term) if self.is_input else (term, self)

        assert not (
            output_term.node_type == "actions" and input_term.node_type == "observations"
        ), "Cannot connect actions directly to observations."

        source = output_term.connection_view()
        target = input_term.connection_view()

        space_converter = None

        observation = target()[2] if target()[0] == "env/observations" else None
        action = source()[2] if source()[0] == "env/actions" else None
        target_params = target.to_dict()
        source_params = source.to_dict()
        identity_converter = {"converter_type": get_module_type_string(Identity)}
        if len(target_params) == 0:
            # If the target is an observation, we need to get the space_converter from the source
            converter = source_params["space_converter"] if "space_converter" in source_params else identity_converter
            converter = converter if converter is not None else identity_converter
            delay, window = 0.0, 0
        else:
            if len(source_params) == 0:
                # If action has no converter, we first open a dialog in which the user can modify the space converter
                parent = self.node.graph.widget().cwWin
                library = self.node.graph.library
                msg_type_out = None
                if "converter" in target_params and target_params["converter"] is not None:
                    msg_type_in = get_opposite_msg_cls(target_params["msg_type"], target_params["converter"])
                else:
                    msg_type_in = get_attribute_from_module(target_params["msg_type"])
                space_converter = target_params["space_converter"] if target_params["space_converter"] else identity_converter
                space_converter_dialog = ConverterDialog(
                    space_converter,
                    parent,
                    library,
                    msg_type_in,
                    msg_type_out,
                    is_space_converter=True,
                )
                space_converter_dialog.setWindowTitle("Space Converter {}".format(output_term.terminal_name))
                space_converter = space_converter_dialog.open()
            delay = target_params["delay"] if "delay" in target_params else None
            window = target_params["window"] if "window" in target_params else None
            converter = target_params["converter"] if "converter" in target_params else identity_converter
            converter = converter if converter is not None else identity_converter
        connect_params = connection_item.open_connection_dialog(converter=converter, delay=delay, window=window)
        target = None if observation else target
        source = None if action else source

        # Convert source and target to new API (i.e. use Lookup)
        self.node.graph.connect(
            source=source,
            target=target,
            action=action,
            observation=observation,
            **connect_params,
        )

        if space_converter is not None:
            output_term.set_param("converter", space_converter)

        self.connections[term] = connection_item
        term.connections[self] = connection_item

        self.connected(term)
        term.connected(self)

        return connection_item

    def disconnect_from(self, term):
        if not self.connected_to(term):
            return
        item = self.connections[term]
        item.close()
        del self.connections[term]
        del term.connections[self]

        self.disconnected(term)
        term.disconnected(self)

    def disconnect_all(self):
        for t in list(self.connections.keys()):
            self.disconnect_from(t)

    def recolor(self, color=None, recurse=True):
        if color is None:
            if self.is_state:
                if self.is_connectable:
                    color = QtGui.QColor(255, 0, 0)
                else:
                    color = QtGui.QColor(128, 128, 128)
            elif self.is_feedthrough:
                color = QtGui.QColor(175, 238, 238)
            else:
                color = QtGui.QColor(0, 0, 255)
        self.graphics_item().setBrush(QtGui.QBrush(color))

        if recurse:
            for t in self.connections:
                t.recolor(color, recurse=False)

    def rename(self, name):
        if self.is_input:
            if name in self.node.inputs:
                self.graphics_item().term_renamed(self.name)
                return
        elif name in self.node.outputs:
            self.graphics_item().term_renamed(self.name)
            return
        node_name, _, old_name = self.connection_tuple()
        if node_name == "env/actions":
            self.node.graph.rename(new=name.split("/")[-1], action=old_name)
        elif node_name == "env/observations":
            self.node.graph.rename(new=name.split("/")[-1], observation=old_name)
        else:
            raise NotImplementedError("Can only change terminal names for actions and observations.")
        old_name = self.name
        self.name = name
        self.terminal_name = name.split("/")[-1]
        self.node.terminal_renamed(self, old_name)
        self.graphics_item().term_renamed(name)

    def __repr__(self):
        return "<Terminal %s.%s>" % (str(self.node.name), str(self.terminal_name))

    def __hash__(self):
        return id(self)

    def close(self):
        self.disconnect_all()
        item = self.graphics_item()
        if item.scene() is not None:
            item.scene().removeItem(item)


class TextItem(QtWidgets.QGraphicsTextItem):
    def __init__(self, text, parent, on_update):
        super().__init__(text, parent)
        self.on_update = on_update

    def focusOutEvent(self, ev):
        super().focusOutEvent(ev)
        if self.on_update is not None:
            self.on_update()

    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            if self.on_update is not None:
                self.on_update()
                return
        super().keyPressEvent(ev)


class TerminalGraphicsItem(GraphicsObject):
    def __init__(self, term, parent=None):
        self.term = term
        GraphicsObject.__init__(self, parent)
        self.brush = fn.mkBrush(0, 0, 0)
        self.box = QtWidgets.QGraphicsRectItem(0, 0, 10, 10, self)
        on_update = self.label_changed if self.term.is_renamable else None
        self.label = TextItem(self.term.terminal_name, self, on_update)
        self.label.setScale(0.7)
        self.newConnection = None
        self.setFiltersChildEvents(True)  # to pick up mouse events on the rectitem
        if self.term.is_renamable:
            self.label.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextEditorInteraction)
            self.label.focusOutEvent = self.label_focus_out
            self.label.keyPressEvent = self.label_key_press
        self.setZValue(1)
        self.menu = None

    def label_focus_out(self, ev):
        QtWidgets.QGraphicsTextItem.focusOutEvent(self.label, ev)
        self.label_changed()

    def label_key_press(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            self.label_changed()
        else:
            QtWidgets.QGraphicsTextItem.keyPressEvent(self.label, ev)

    def label_changed(self):
        new_name = str(self.label.toPlainText())
        if new_name != self.term.terminal_name:
            new_name = self.term.terminal_type + "/" + new_name
            self.term.rename(new_name)

    def term_renamed(self, name):
        self.label.setPlainText(name.split("/")[-1])

    def setBrush(self, brush):
        self.brush = brush
        self.box.setBrush(brush)

    def disconnect(self, target):
        input_term, output_term = (self.term, target.term) if self.term.is_input else (target.term, self.term)
        if output_term.node_type == "actions":
            self.term.node.graph.disconnect(
                action=output_term.terminal_name,
                target=input_term.connection_view(),
                remove=False,
            )
        elif input_term.node_type == "observations":
            self.term.node.graph.disconnect(
                source=output_term.connection_view(),
                observation=input_term.terminal_name,
                remove=False,
            )
        else:
            self.term.node.graph.disconnect(
                source=output_term.connection_view(),
                target=input_term.connection_view(),
            )
        self.term.disconnect_from(target.term)

    def boundingRect(self):
        br = self.box.mapRectToParent(self.box.boundingRect())
        lr = self.label.mapRectToParent(self.label.boundingRect())
        return br | lr

    def paint(self, p, *args):
        pass

    def set_anchor(self, x, y):
        pos = QtCore.QPointF(x, y)
        self.anchorPos = pos
        br = self.box.mapRectToParent(self.box.boundingRect())
        lr = self.label.mapRectToParent(self.label.boundingRect())

        if self.term.is_input:
            self.box.setPos(pos.x(), pos.y() - br.height() / 2.0)
            self.label.setPos(pos.x() + br.width(), pos.y() - lr.height() / 2.0)
        else:
            self.box.setPos(pos.x() - br.width(), pos.y() - br.height() / 2.0)
            self.label.setPos(pos.x() - br.width() - lr.width(), pos.y() - lr.height() / 2.0)
        self.update_connections()

    def update_connections(self):
        for t, c in self.term.connections.items():
            c.update_line()

    def mousePressEvent(self, ev):
        # ev.accept()
        ev.ignore()  # necessary to allow click/drag events to process correctly

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            self.label.setFocus(QtCore.Qt.FocusReason.MouseFocusReason)
        elif ev.button() == QtCore.Qt.MouseButton.RightButton:
            ev.accept()
            self.raise_context_menu(ev)

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            param_window = ParamWindow(node=self.term.node, term=self.term)
            param_window.open()
            param_window.close()

    def raise_context_menu(self, ev):
        # only raise menu if this terminal is removable
        menu = self.get_menu()
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))

    def get_menu(self):
        self.menu = QtWidgets.QMenu()
        self.menu.setTitle("Terminal")
        rem_act = QtGui.QAction("Remove {}".format(self.term.terminal_name), self.menu)
        rem_act.triggered.connect(self.remove_self)
        self.menu.addAction(rem_act)
        self.menu.remAct = rem_act
        if not self.term.is_removable:
            rem_act.setEnabled(False)
        return self.menu

    def remove_self(self):
        if self.term.node_type == "actions":
            self.term.node.graph.remove_component(action=self.term.terminal_name)
        elif self.term.node_type == "observations":
            self.term.node.graph.remove_component(observation=self.term.terminal_name)
        else:
            self.term.node.graph.remove_component(self.term.connection_view())
        self.term.node.remove_terminal(self.term)

    def mouseDragEvent(self, ev):
        if ev.button() != QtCore.Qt.MouseButton.LeftButton:
            ev.ignore()
            return
        if not self.term.is_connectable:
            return
        ev.accept()
        if ev.isStart():
            if self.newConnection is None:
                self.newConnection = ConnectionItem(self)
                self.getViewBox().addItem(self.newConnection)

            self.newConnection.set_target(self.mapToView(ev.pos()))
        elif ev.isFinish():
            if self.newConnection is not None:
                items = self.scene().items(ev.scenePos())
                got_target = False
                for i in items:
                    if isinstance(i, TerminalGraphicsItem):
                        self.newConnection.set_target(i)
                        try:
                            self.term.connect_to(
                                i.term,
                                self.newConnection,
                                graph_backup=self.term.node.graph,
                            )
                            got_target = True
                        except Exception:
                            got_target = False
                        break

                if not got_target:
                    self.newConnection.close()
                self.newConnection = None
        else:
            if self.newConnection is not None:
                self.newConnection.set_target(self.mapToView(ev.pos()))

    def hoverEvent(self, ev):
        if not ev.isExit() and ev.acceptDrags(QtCore.Qt.MouseButton.LeftButton):
            ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton)
            # we don't use the click, but we also don't want anyone else to use it.
            ev.acceptClicks(QtCore.Qt.MouseButton.RightButton)
            self.box.setBrush(fn.mkBrush("w"))
        else:
            self.box.setBrush(self.brush)
        self.update()

    def connect_point(self):
        # return the connect position of this terminal in view coords
        return self.mapToView(self.mapFromItem(self.box, self.box.boundingRect().center()))

    def node_moved(self):
        for t, item in self.term.connections.items():
            item.update_line()


class ConnectionItem(GraphicsObject):
    def __init__(self, source, target=None):
        GraphicsObject.__init__(self)
        self.setFlags(self.GraphicsItemFlag.ItemIsSelectable | self.GraphicsItemFlag.ItemIsFocusable)
        self.source = source
        self.target = target
        self.length = 0
        self.hovered = False
        self.path = None
        self.shapePath = None
        self.connection_window = None

        if self.source.term.is_state:
            self.style = {
                "shape": "line",
                "color": (255, 0, 0, 255),
                "width": 2.0,
                "hoverColor": (150, 150, 250, 255),
                "hoverWidth": 2.0,
                "selectedColor": (200, 200, 0, 255),
                "selectedWidth": 4.0,
            }
        else:
            self.style = {
                "shape": "line",
                "color": (0, 0, 255, 255),
                "width": 2.0,
                "hoverColor": (150, 150, 250),
                "hoverWidth": 2.0,
                "selectedColor": (200, 200, 0, 255),
                "selectedWidth": 4.0,
            }
        self.source.getViewBox().addItem(self)
        self.update_line()

    def open_connection_dialog(self, **kwargs):
        input_term, output_term = (
            (self.source.term, self.target.term) if self.source.term.is_input else (self.target.term, self.source.term)
        )
        self.connection_window = ConnectionDialog(input_term=input_term, output_term=output_term, **kwargs)
        return self.connection_window.open()

    def close(self):
        if self.scene() is not None:
            self.scene().removeItem(self)

    def set_target(self, target):
        self.target = target
        self.update_line()

    def setStyle(self, **kwds):
        self.style.update(kwds)
        if "shape" in kwds:
            self.update_line()
        else:
            self.update()

    def update_line(self):
        start = Point(self.source.connect_point())
        if isinstance(self.target, TerminalGraphicsItem):
            stop = Point(self.target.connect_point())
            input_term = self.target.term if self.target.term.is_input else self.source.term
            node_name = input_term.node.name
            linestyle_state = input_term.node.graph._state["gui_state"][node_name]["linestyle"]
            if input_term.name in linestyle_state:
                shape = linestyle_state[input_term.name]
            else:
                shape = "cubic"
                linestyle_state[input_term.name] = shape
            self.style["shape"] = shape
        elif isinstance(self.target, QtCore.QPointF):
            stop = Point(self.target)
        else:
            return
        self.prepareGeometryChange()

        self.path = self.generate_path(start, stop)
        self.shapePath = None
        self.update()
        self.setZValue(-1)

    def generate_path(self, start, stop):
        path = QtGui.QPainterPath()
        path.moveTo(start)
        if self.style["shape"] == "line":
            path.lineTo(stop)
        elif self.style["shape"] == "cubic":
            path.cubicTo(
                Point(stop.x(), start.y()),
                Point(start.x(), stop.y()),
                Point(stop.x(), stop.y()),
            )
        else:
            raise Exception('Invalid shape "%s"; options are "line" or "cubic"' % self.style["shape"])
        return path

    def keyPressEvent(self, ev):
        if not self.isSelected():
            ev.ignore()
            return

        if ev.key() == QtCore.Qt.Key.Key_Delete or ev.key() == QtCore.Qt.Key.Key_Backspace:
            if self.source.term.node_type == "engine_node":
                ev.ignore()
                return
            self.source.disconnect(self.target)
            ev.accept()
            if isinstance(self.target, TerminalGraphicsItem):
                # Remove linestyle from gui state on removal
                input_term = self.target.term if self.target.term.is_input else self.source.term
                node_name = input_term.node.name
                if input_term.name in input_term.node.graph._state["gui_state"][node_name]["linestyle"]:
                    input_term.node.graph._state["gui_state"][node_name]["linestyle"].pop(input_term.name)
        elif ev.key() == QtCore.Qt.Key.Key_Control:
            ev.accept()
            if self.style["shape"] == "line":
                shape = "cubic"
            elif self.style["shape"] == "cubic":
                shape = "line"
            if isinstance(self.target, TerminalGraphicsItem):
                # Update linestyle in gui state
                input_term = self.target.term if self.target.term.is_input else self.source.term
                node_name = input_term.node.name
                input_term.node.graph._state["gui_state"][node_name]["linestyle"][input_term.name] = shape
                self.setStyle(shape=shape)
        else:
            ev.ignore()

    def mousePressEvent(self, ev):
        ev.ignore()

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            sel = self.isSelected()
            self.setSelected(True)
            self.setFocus()
            if not sel and self.isSelected():
                self.update()

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            if self.connection_window is not None:
                self.connection_window.show()

    def hoverEvent(self, ev):
        if (not ev.isExit()) and ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton):
            self.hovered = True
        else:
            self.hovered = False
        self.update()

    def boundingRect(self):
        return self.shape().boundingRect()

    def viewRangeChanged(self):
        self.shapePath = None
        self.prepareGeometryChange()

    def shape(self):
        if self.shapePath is None:
            if self.path is None:
                return QtGui.QPainterPath()
            stroker = QtGui.QPainterPathStroker()
            px = self.pixelWidth()
            stroker.setWidth(px * 8)
            self.shapePath = stroker.createStroke(self.path)
        return self.shapePath

    def paint(self, p, *args):
        if self.isSelected():
            p.setPen(fn.mkPen(self.style["selectedColor"], width=self.style["selectedWidth"]))
        else:
            if self.hovered:
                p.setPen(fn.mkPen(self.style["hoverColor"], width=self.style["hoverWidth"]))
            else:
                p.setPen(fn.mkPen(self.style["color"], width=self.style["width"]))

        p.drawPath(self.path)



"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Flowchart.py
"""

# -*- coding: utf-8 -*-
import os
import importlib
from copy import deepcopy
from functools import partial

# Import pyqtgraph modules
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph.Qt import QT_LIB, QtCore, QtWidgets
from pyqtgraph import FileDialog, DataTreeWidget
from pyqtgraph import dockarea
from pyqtgraph.debug import printExc

# eagerx imports
from eagerx.core.graph import Graph
from eagerx.core.graph_engine import EngineGraph
from eagerx.core.entities import Node

# eagerx_gui imports
from eagerx_gui import configuration
from eagerx_gui.utils import get_nodes_and_objects_library, add_pos_to_state, add_pos_to_engine_state, empty_gui_state
from eagerx_gui import gui_view
from eagerx_gui.gui_node import RxGuiNode, RxEngineGuiNode, NodeGraphicsItem
from eagerx_gui.gui_terminal import TerminalGraphicsItem, ConnectionItem
from eagerx_gui.pyqtgraph_utils import exception_handler, NodeCreationDialog

# pyside and pyqt use incompatible ui files.
rx_ui_template = importlib.import_module("eagerx_gui.templates.ui_{}".format(QT_LIB.lower()))
rx_engine_ui_template = importlib.import_module("eagerx_gui.templates.engine_ui_{}".format(QT_LIB.lower()))


class EngineGui(EngineGraph, QtCore.QObject):
    library = get_nodes_and_objects_library()
    sigFileLoaded = QtCore.Signal(object)
    sigFileSaved = QtCore.Signal(object)
    sigChartLoaded = QtCore.Signal()
    sigStateChanged = QtCore.Signal()  # called when output is expected to have changed
    sigChartChanged = QtCore.Signal(object, object, object)  # called when nodes are added, removed, or renamed.

    def __init__(self, state):
        EngineGraph.__init__(self, state=state)
        QtCore.QObject.__init__(self)
        self.nodes = {}
        self.next_z_val = 10
        self._widget = None
        self.scene = None
        self.file_path = None
        self.widget()
        self.load_state(clear=True)
        self.viewBox.autoRange(padding=0.04)

    def create_node(self, name, pos):
        """Create a new Node and add it to this flowchart."""
        node = RxEngineGuiNode(name, graph=self)
        self.add_node(node, name, pos)
        return node

    def add_node(self, node, name, pos=None):
        """Add an existing Node to this flowchart.

        See also: createnode
        """
        if pos is None:
            pos = [0, 0]
        if type(pos) in [QtCore.QPoint, QtCore.QPointF]:
            pos = [pos.x(), pos.y()]
        item = node.graphics_item()
        item.setZValue(self.next_z_val * 2)
        self.next_z_val += 1
        self.viewBox.addItem(item)
        item.moveBy(*pos)
        self.nodes[name] = node
        self.widget().add_node(node)
        node.sigClosed.connect(self.node_closed)
        node.sigRenamed.connect(self.node_renamed)
        self.sigChartChanged.emit(self, "add", node)

    def remove_node(self, node):
        """Remove a Node from this flowchart."""
        node.close()

    def node_closed(self, node):
        del self.nodes[node.name]
        self.widget().remove_node(node)
        for signal in ["sigClosed", "sigRenamed"]:
            try:
                getattr(node, signal).disconnect(self.node_closed)
            except (TypeError, RuntimeError):
                pass
        self.sigChartChanged.emit(self, "remove", node)

    def node_renamed(self, node, old_name):
        NotImplementedError("You cannot rename entities once added.")

    def connect_terminals(self, term1, term2):
        """Connect two terminals together within this flowchart."""
        connection_item = ConnectionItem(term1.graphics_item(), term2.graphics_item())
        term1.graphics_item().getViewBox().addItem(connection_item)

        term1.connections[term2] = connection_item
        term2.connections[term1] = connection_item

        term1.connected(term2)
        term2.connected(term1)

    def chart_graphics_item(self):
        """Return the graphicsItem that displays the internal nodes and
        connections of this flowchart.

        Note that the similar method `graphicsItem()` is inherited from Node
        and returns the *external* graphical representation of this flowchart."""
        return self.viewBox

    def widget(self):
        """Return the control widget for this flowchart.

        This widget provides GUI access to the parameters for each node and a
        graphical representation of the flowchart.
        """
        if self._widget is None:
            self._widget = EngineCtrlWidget(self)
            self.scene = self._widget.scene
            self.viewBox = self._widget.viewBox()
        return self._widget

    def load_state(self, clear=False):
        self.blockSignals(True)
        try:
            if clear:
                self.clear()
            add_pos_to_engine_state(self._state)
            nodes = self._state["nodes"]
            nodes = [dict(**node, name=name, gui_state=self._state["gui_state"][name]) for name, node in nodes.items()]
            nodes.sort(key=lambda a: a["gui_state"]["pos"][0])
            for n in nodes:
                if n["name"] in self.nodes:
                    self.nodes[n["name"]].load_state(n)
                    continue
                try:
                    name = n["name"]
                    pos = n["gui_state"]["pos"]
                    node = self.create_node(name, pos)
                    node.load_state(n)
                except Exception:
                    printExc("Error creating node %s: (continuing anyway)" % n["name"])

            connects = [
                (
                    connection[0][0],
                    "/".join(connection[0][1:3]),
                    connection[1][0],
                    "/".join(connection[1][1:3]),
                )
                for connection in self._state["connects"]
            ]
            for n1, t1, n2, t2 in connects:
                try:
                    self.connect_terminals(self.nodes[n1].terminals[t1], self.nodes[n2].terminals[t2])
                except Exception:
                    print(self.nodes[n1].terminals)
                    print(self.nodes[n2].terminals)
                    printExc("Error connecting terminals %s.%s - %s.%s:" % (n1, t1, n2, t2))

        finally:
            self.blockSignals(False)

        self.sigChartLoaded.emit()
        self.sigStateChanged.emit()

    def state(self):
        return deepcopy(self._state)

    def clear(self):
        """Remove all nodes from this flowchart except the original input/output nodes."""
        for n in list(self.nodes.values()):
            n.close()  # calls self.nodeClosed(n) by signal
        self.widget().clear()

    def tuple_to_view(self, t):
        if isinstance(t, tuple):
            t = list(t)
        return self.get_view(t[0], depth=t[1:])


class Gui(Graph, QtCore.QObject):
    library = get_nodes_and_objects_library()
    sigFileLoaded = QtCore.Signal(object)
    sigFileSaved = QtCore.Signal(object)
    sigChartLoaded = QtCore.Signal()
    sigStateChanged = QtCore.Signal()  # called when output is expected to have changed
    sigChartChanged = QtCore.Signal(object, object, object)  # called when nodes are added, removed, or renamed.

    def __init__(self, state):
        Graph.__init__(self, state=state)
        QtCore.QObject.__init__(self)
        self.nodes = {}
        self.next_z_val = 10
        self._widget = None
        self.scene = None
        self.file_path = None
        self.widget()
        self.load_state(clear=True)
        self.viewBox.autoRange(padding=0.04)

    def create_node(self, name, pos):
        """Create a new Node and add it to this flowchart."""
        node = RxGuiNode(name, graph=self)
        self.add_node(node, name, pos)
        return node

    def add_node(self, node, name, pos=None):
        """Add an existing Node to this flowchart.

        See also: createnode
        """
        if pos is None:
            pos = [0, 0]
        if type(pos) in [QtCore.QPoint, QtCore.QPointF]:
            pos = [pos.x(), pos.y()]
        item = node.graphics_item()
        item.setZValue(self.next_z_val * 2)
        self.next_z_val += 1
        self.viewBox.addItem(item)
        item.moveBy(*pos)
        self.nodes[name] = node
        self.widget().add_node(node)
        node.sigClosed.connect(self.node_closed)
        node.sigRenamed.connect(self.node_renamed)
        self.sigChartChanged.emit(self, "add", node)

    def remove_node(self, node):
        """Remove a Node from this flowchart."""
        node.close()

    def node_closed(self, node):
        del self.nodes[node.name]
        self.widget().remove_node(node)
        for signal in ["sigClosed", "sigRenamed"]:
            try:
                getattr(node, signal).disconnect(self.node_closed)
            except (TypeError, RuntimeError):
                pass
        self.sigChartChanged.emit(self, "remove", node)

    def node_renamed(self, node, old_name):
        NotImplementedError("You cannot rename entities once added.")
        # del self.nodes[old_name]
        # self.nodes[node.name] = node
        # self.widget().node_renamed(node, old_name)
        # self.sigChartChanged.emit(self, "rename", node)

    def connect_terminals(self, term1, term2):
        """Connect two terminals together within this flowchart."""
        connection_item = ConnectionItem(term1.graphics_item(), term2.graphics_item())
        term1.graphics_item().getViewBox().addItem(connection_item)

        term1.connections[term2] = connection_item
        term2.connections[term1] = connection_item

        term1.connected(term2)
        term2.connected(term1)

    def chart_graphics_item(self):
        """Return the graphicsItem that displays the internal nodes and
        connections of this flowchart.

        Note that the similar method `graphicsItem()` is inherited from Node
        and returns the *external* graphical representation of this flowchart."""
        return self.viewBox

    def widget(self):
        """Return the control widget for this flowchart.

        This widget provides GUI access to the parameters for each node and a
        graphical representation of the flowchart.
        """
        if self._widget is None:
            self._widget = CtrlWidget(self)
            self.scene = self._widget.scene
            self.viewBox = self._widget.viewBox()
        return self._widget

    def load_state(self, clear=False):
        self.blockSignals(True)
        try:
            if clear:
                self.clear()
            if "env/render" not in self._state["nodes"]:
                render = Node.make("Render", rate=1.0)
                self.add(render)
            add_pos_to_state(self._state)
            nodes = self._state["nodes"]
            nodes = [dict(**node, name=name, gui_state=self._state["gui_state"][name]) for name, node in nodes.items()]
            nodes.sort(key=lambda a: a["gui_state"]["pos"][0])
            for n in nodes:
                if n["name"] in self.nodes:
                    self.nodes[n["name"]].load_state(n)
                    continue
                try:
                    name = n["name"]
                    pos = n["gui_state"]["pos"]
                    node = self.create_node(name, pos)
                    node.load_state(n)
                except Exception:
                    printExc("Error creating node %s: (continuing anyway)" % n["name"])

            connects = [
                (
                    connection[0][0],
                    "/".join(connection[0][1:3]),
                    connection[1][0],
                    "/".join(connection[1][1:3]),
                )
                for connection in self._state["connects"]
            ]
            for n1, t1, n2, t2 in connects:
                try:
                    self.connect_terminals(self.nodes[n1].terminals[t1], self.nodes[n2].terminals[t2])
                except Exception:
                    print(self.nodes[n1].terminals)
                    print(self.nodes[n2].terminals)
                    printExc("Error connecting terminals %s.%s - %s.%s:" % (n1, t1, n2, t2))

        finally:
            self.blockSignals(False)

        self.sigChartLoaded.emit()
        self.sigStateChanged.emit()

    def state(self):
        state = deepcopy(self._state)
        target = ["env/render", "inputs", "image"]
        render_connected = False
        for idx, c in enumerate(self._state["connects"]):
            if target == c[1]:
                render_connected = True
                break
        if not render_connected:
            state["nodes"].pop("env/render")
        return state

    def load_file(self, file_name=None, start_dir=None):
        """Load a flowchart (``*.graph``) file."""
        if file_name is None:
            if start_dir is None:
                start_dir = self.file_path
            if start_dir is None:
                start_dir = ""
            self.fileDialog = FileDialog(None, "Load Graph..", start_dir, "Graph (*.graph)")
            self.fileDialog.show()
            self.fileDialog.fileSelected.connect(self.load_file)
            return
        file_name = file_name
        self.load(file_name)
        self.load_state(clear=True)
        self.viewBox.autoRange()
        self.sigFileLoaded.emit(file_name)

    def save_file(self, file_name=None, start_dir=None):
        """Save this flowchart to a .graph file"""
        if file_name is None:
            if start_dir is None:
                start_dir = self.file_path
            if start_dir is None:
                start_dir = ""
            self.fileDialog = FileDialog(None, "Save Graph..", start_dir, "Graph (*.graph)")
            self.fileDialog.setDefaultSuffix("graph")
            self.fileDialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptSave)
            self.fileDialog.show()
            self.fileDialog.fileSelected.connect(self.save_file)
            return
        file_name = file_name
        self._state = self.state()
        self.save(file_name)
        self.sigFileSaved.emit(file_name)
        self.load_state(clear=True)

    def clear(self):
        """Remove all nodes from this flowchart except the original input/output nodes."""
        for n in list(self.nodes.values()):
            n.close()  # calls self.nodeClosed(n) by signal
        self.widget().clear()

    def tuple_to_view(self, t):
        if isinstance(t, tuple):
            t = list(t)
        return self.get_view(t[0], depth=t[1:])


class GraphicsItem(GraphicsObject):
    def __init__(self, chart):
        GraphicsObject.__init__(self)
        self.chart = chart  # chart is an instance of Flowchart()
        self.update_terminals()

    def update_terminals(self):
        self.terminals = {}

    def boundingRect(self):
        return QtCore.QRectF()

    def paint(self, p, *args):
        pass


class EngineCtrlWidget(QtWidgets.QWidget):
    """The widget that contains the list of all the nodes in a flowchart and their controls, as well as buttons for
    loading/saving flowcharts."""

    def __init__(self, chart):
        self.items = {}
        self.current_file_name = None
        QtWidgets.QWidget.__init__(self)
        self.chart = chart
        self.ui = rx_engine_ui_template.Ui_Form()
        self.ui.setupUi(self)
        self.ui.ctrlList.setColumnCount(2)
        self.ui.ctrlList.setColumnWidth(1, 20)
        self.ui.ctrlList.setVerticalScrollMode(self.ui.ctrlList.ScrollMode.ScrollPerPixel)
        self.ui.ctrlList.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.chartWidget = EagerxGraphWidget(chart, self)
        self.scene = self.chartWidget.scene
        self.cwWin = QtWidgets.QMainWindow()
        self.cwWin.setWindowTitle("EAGERx EngineGraph")
        self.cwWin.setCentralWidget(self.chartWidget)
        self.cwWin.resize(1000, 800)

        h = self.ui.ctrlList.header()
        h.setSectionResizeMode(0, h.ResizeMode.Stretch)

        self.ui.ctrlList.itemChanged.connect(self.itemChanged)
        self.ui.showChartBtn.toggled.connect(self.chart_toggled)
        self.ui.checkValidityBtn.clicked.connect(self.check_validity_toggled)

        self.chart.sigFileLoaded.connect(self.set_current_file)

    def chart_toggled(self, b):
        if b:
            self.cwWin.show()
        else:
            self.cwWin.hide()

    def check_validity_toggled(self):
        try:
            self._check_validity(graph_backup=self.chart, dialog_title="Invalid Graph")
            self.ui.checkValidityBtn.success("Valid")
        except Exception:
            self.ui.checkValidityBtn.failure("Invalid")

    @exception_handler
    def _check_validity(self):
        self.chart._is_valid(state=self.chart.state())

    def set_current_file(self, file_name):
        self.current_file_name = file_name
        if file_name is None:
            self.ui.fileNameLabel.setText("<b>[ new ]</b>")
        else:
            self.ui.fileNameLabel.setText("<b>%s</b>" % os.path.split(self.current_file_name)[1])
        self.resizeEvent(None)

    def itemChanged(self, *args):
        pass

    def viewBox(self):
        return self.chartWidget.viewBox()

    def add_node(self, node):
        ctrl = node.ctrl_widget()
        item = QtWidgets.QTreeWidgetItem([node.name, "", ""])
        self.ui.ctrlList.addTopLevelItem(item)

        if ctrl is not None:
            item2 = QtWidgets.QTreeWidgetItem()
            item.addChild(item2)
            self.ui.ctrlList.setItemWidget(item2, 0, ctrl)

        self.items[node] = item

    def remove_node(self, node):
        if node in self.items:
            item = self.items[node]
            self.ui.ctrlList.removeTopLevelItem(item)

    def clear(self):
        self.chartWidget.clear()

    def select(self, node):
        item = self.items[node]
        self.ui.ctrlList.setCurrentItem(item)

    def clearSelection(self):
        self.ui.ctrlList.selectionModel().clearSelection()


class CtrlWidget(QtWidgets.QWidget):
    """The widget that contains the list of all the nodes in a flowchart and their controls, as well as buttons for
    loading/saving flowcharts."""

    def __init__(self, chart):
        self.items = {}
        self.current_file_name = None
        QtWidgets.QWidget.__init__(self)
        self.chart = chart
        self.ui = rx_ui_template.Ui_Form()
        self.ui.setupUi(self)
        self.ui.ctrlList.setColumnCount(2)
        self.ui.ctrlList.setColumnWidth(1, 20)
        self.ui.ctrlList.setVerticalScrollMode(self.ui.ctrlList.ScrollMode.ScrollPerPixel)
        self.ui.ctrlList.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.chartWidget = EagerxGraphWidget(chart, self)
        self.scene = self.chartWidget.scene
        self.cwWin = QtWidgets.QMainWindow()
        self.cwWin.setWindowTitle("EAGERx Graph")
        self.cwWin.setCentralWidget(self.chartWidget)
        self.cwWin.resize(1000, 800)

        h = self.ui.ctrlList.header()
        h.setSectionResizeMode(0, h.ResizeMode.Stretch)

        self.ui.ctrlList.itemChanged.connect(self.itemChanged)
        self.ui.loadBtn.clicked.connect(self.load_clicked)
        self.ui.saveBtn.clicked.connect(self.save_clicked)
        self.ui.saveAsBtn.clicked.connect(self.save_as_clicked)
        self.ui.showChartBtn.toggled.connect(self.chart_toggled)
        self.ui.checkValidityBtn.clicked.connect(self.check_validity_toggled)
        # self.ui.showCompatibleBridgesBtn.clicked.connect(self.show_compatible_bridges_toggled)

        self.chart.sigFileLoaded.connect(self.set_current_file)
        self.chart.sigFileSaved.connect(self.file_saved)

    def chart_toggled(self, b):
        if b:
            self.cwWin.show()
        else:
            self.cwWin.hide()

    def load_clicked(self):
        new_file = self.chart.load_file()

    def file_saved(self, file_name):
        self.set_current_file(file_name)
        self.ui.saveBtn.success("Saved.")

    def save_clicked(self):
        if self.current_file_name is None:
            self.save_as_clicked()
        else:
            try:
                self.chart.save_file(self.current_file_name)
            except:
                self.ui.saveBtn.failure("Error")
                raise

    def save_as_clicked(self):
        try:
            if self.current_file_name is None:
                new_file = self.chart.save_file()
            else:
                new_file = self.chart.save_file(suggested_file_name=self.current_file_name)
        except:
            self.ui.saveBtn.failure("Error")
            raise

    def check_validity_toggled(self):
        try:
            self._check_validity(graph_backup=self.chart, dialog_title="Invalid Graph")
            self.ui.checkValidityBtn.success("Valid")
        except Exception:
            self.ui.checkValidityBtn.failure("Invalid")

    @exception_handler
    def _check_validity(self):
        self.chart._is_valid(state=self.chart.state())

    # def show_compatible_bridges_toggled(self):
    #     try:
    #         label_string = self.chart.check_exists_compatible_bridge(self.chart.state(), tablefmt="html")
    #     except Exception as e:
    #         label_string = str(e)
    #     bridges_window = QtWidgets.QDialog(self.chart.widget().cwWin)
    #     bridges_window.setWindowTitle("Compatible Bridges")
    #     layout = QtWidgets.QGridLayout()
    #     label = QtWidgets.QLabel(label_string)
    #     label.setWordWrap(True)
    #     layout.addWidget(label)
    #     bridges_window.setLayout(layout)
    #     bridges_window.exec()

    def set_current_file(self, file_name):
        self.current_file_name = file_name
        if file_name is None:
            self.ui.fileNameLabel.setText("<b>[ new ]</b>")
        else:
            self.ui.fileNameLabel.setText("<b>%s</b>" % os.path.split(self.current_file_name)[1])
        self.resizeEvent(None)

    def itemChanged(self, *args):
        pass

    def viewBox(self):
        return self.chartWidget.viewBox()

    def node_renamed(self, node, old_name):
        self.items[node].setText(0, node.name)

    def add_node(self, node):
        ctrl = node.ctrl_widget()
        item = QtWidgets.QTreeWidgetItem([node.name, "", ""])
        self.ui.ctrlList.addTopLevelItem(item)

        if ctrl is not None:
            item2 = QtWidgets.QTreeWidgetItem()
            item.addChild(item2)
            self.ui.ctrlList.setItemWidget(item2, 0, ctrl)

        self.items[node] = item

    def remove_node(self, node):
        if node in self.items:
            item = self.items[node]
            self.ui.ctrlList.removeTopLevelItem(item)

    def clear(self):
        self.chartWidget.clear()

    def select(self, node):
        item = self.items[node]
        self.ui.ctrlList.setCurrentItem(item)

    def clearSelection(self):
        self.ui.ctrlList.selectionModel().clearSelection()


class EagerxGraphWidget(dockarea.DockArea):
    """Includes the actual graphical flowchart and debugging interface"""

    def __init__(self, chart, ctrl):
        dockarea.DockArea.__init__(self)
        self.chart = chart
        self.ctrl = ctrl
        self.hoverItem = None

        # build user interface (it was easier to do it here than via developer)
        self.view = gui_view.RxView(self)
        self.viewDock = dockarea.Dock("view", size=(1000, 600))
        self.viewDock.addWidget(self.view)
        self.viewDock.hideTitleBar()
        self.addDock(self.viewDock)

        self.hoverText = QtWidgets.QTextEdit()
        self.hoverText.setReadOnly(True)
        self.hoverDock = dockarea.Dock("Hover Info", size=(1000, 20))
        self.hoverDock.addWidget(self.hoverText)
        self.addDock(self.hoverDock, "bottom")

        self.selInfo = QtWidgets.QWidget()
        self.selInfoLayout = QtWidgets.QGridLayout()
        self.selInfo.setLayout(self.selInfoLayout)
        self.selDescLabel = QtWidgets.QLabel()
        self.selNameLabel = QtWidgets.QLabel()
        self.selDescLabel.setWordWrap(True)
        self.selectedTree = DataTreeWidget()
        self.selInfoLayout.addWidget(self.selDescLabel)
        self.selInfoLayout.addWidget(self.selectedTree)

        self.scene = self.view.scene()
        self._viewBox = self.view.viewBox()

        self.buildMenu()

        self.scene.selectionChanged.connect(self.selection_changed)
        self.scene.sigMouseHover.connect(self.hover_over)

    def buildMenu(self, pos=None):
        def build_sub_menu(library, root_menu, submenus, pos=None):
            for node in library:
                id = node["entity_id"]
                if id in configuration.GUI_NODE_IDS_TO_IGNORE:
                    continue
                act = root_menu.addAction(id)
                act.nodeType = node
                act.pos = pos

        self.submenus = []
        self.nodeMenu = []
        for node_type, library in self.chart.library.items():
            if node_type in configuration.GUI_ENTITIES_TO_IGNORE:
                continue
            menu = QtWidgets.QMenu("Add {}".format(node_type.replace("_", " ")))
            build_sub_menu(library, menu, self.submenus, pos=pos)
            menu.triggered.connect(partial(self.node_menu_triggered))
            self.nodeMenu.append(menu)
        return self.nodeMenu

    def menuPosChanged(self, pos):
        self.menuPos = pos

    def showViewMenu(self, ev):
        self.buildMenu(ev.scenePos())
        self.nodeMenu.popup(ev.screenPos())

    def viewBox(self):
        return self._viewBox  # the viewBox that items should be added to

    def node_menu_triggered(self, action):
        try:
            self._node_menu_triggered(action, graph_backup=self.chart)
        except Exception:
            pass

    @exception_handler
    def _node_menu_triggered(self, action):
        node_type = action.nodeType
        if action.pos is not None:
            pos = action.pos
        else:
            pos = self.menuPos
        pos = self.viewBox().mapSceneToView(pos)

        if type(pos) in [QtCore.QPoint, QtCore.QPointF]:
            pos = [pos.x(), pos.y()]

        n = 0
        while True:
            name = "{}".format(node_type["entity_id"])
            if n > 0:
                name += "_{}".format(n + 1)
            if name not in self.chart.nodes:
                break
            n += 1

        node_creation_dialog = NodeCreationDialog(name, node_type, self.chart.widget().cwWin)
        mapping = node_creation_dialog.open()
        if "name" in mapping:
            name = mapping["name"]
        rx_entity = node_type["spec"](**mapping)
        self.chart.add(rx_entity)
        self.chart._state["gui_state"][name] = empty_gui_state()
        self.chart._state["gui_state"][name]["pos"] = pos
        self.chart.create_node(name, pos)

    def selection_changed(self):
        items = self.scene.selectedItems()
        if len(items) == 0:
            data = None
        else:
            item = items[0]
            if hasattr(item, "node") and isinstance(item.node, RxGuiNode):
                n = item.node
                if n in self.ctrl.items:
                    self.ctrl.select(n)
                else:
                    self.ctrl.clearSelection()
                data = {}
                self.selNameLabel.setText(n.name)
                if hasattr(n, "name"):
                    self.selDescLabel.setText("<b>%s</b>: %s" % (n.name, n.__class__.__doc__))
                else:
                    self.selDescLabel.setText("")
                if n.exception is not None:
                    data["exception"] = n.exception
            else:
                data = None
        self.selectedTree.setData(data, hideRoot=True)

    def hover_over(self, items):
        for item in items:
            self.hoverItem = item
            if hasattr(item, "term") and isinstance(item, TerminalGraphicsItem):
                text = "name: " + item.term.terminal_name
                for key, value in item.term.params().items():
                    if key in configuration.GUI_WIDGETS["term"]["hide"]["all"]:
                        continue
                    elif (
                        item.term.node_type in configuration.GUI_WIDGETS["term"]["hide"]
                        and key in configuration.GUI_WIDGETS["term"]["hide"][item.term.node_type]
                    ):
                        continue
                    elif (
                        item.term.terminal_type in configuration.GUI_WIDGETS["term"]["hide"]
                        and key in configuration.GUI_WIDGETS["term"]["hide"][item.term.terminal_type]
                    ):
                        continue
                    text += "\n" + "{}: {}".format(key, value)
                self.hoverText.setPlainText(text)
                return
            elif hasattr(item, "node") and isinstance(item, NodeGraphicsItem):
                text = ""
                for key, value in item.node.params().items():
                    if key in configuration.GUI_WIDGETS["node"]["hide"]["all"]:
                        continue
                    elif (
                        item.node.node_type in configuration.GUI_WIDGETS["node"]["hide"]
                        and key in configuration.GUI_WIDGETS["node"]["hide"][item.node.node_type]
                    ):
                        continue
                    text += "{}: {}\n".format(key, value)
                self.hoverText.setPlainText(text)
                return
        self.hoverText.setPlainText("")

    def clear(self):
        self.selectedTree.setData(None)
        self.hoverText.setPlainText("")
        self.selNameLabel.setText("")
        self.selDescLabel.setText("")



from copy import deepcopy

# Terminal types
TERMS = {
    "object": {"in": {"actuators"}, "out": {"sensors", "states"}},
    "node": {"in": {"inputs"}, "out": {"outputs", "states"}},
    "reset_node": {
        "in": {"feedthroughs", "inputs", "targets"},
        "out": {"outputs", "states"},
    },
}
TERMS_IN = set().union(*[TERMS[key]["in"] for key in TERMS])
TERMS_OUT = set().union(*[TERMS[key]["out"] for key in TERMS])

# Possible entries in GUI
GUI_WIDGETS = {
    "node": {
        "hide": {
            "all": [],
            "actions": ["rate"],
            "observations": ["rate"],
        },
        "items": {
            "color": [
                "black",
                "red",
                "green",
                "yellow",
                "blue",
                "magenta",
                "cyan",
                "white",
                "grey",
            ],
            "log_level": {
                "silent": 0,
                "debug": 10,
                "info": 20,
                "warn": 30,
                "error": 40,
                "fatal": 50,
            },
            "log_level_memory": {
                "silent": 0,
                "debug": 10,
                "info": 20,
                "warn": 30,
                "error": 40,
                "fatal": 50,
            },
            "process": {"new process": 0, "environment": 1, "bridge": 2, "external": 3},
        },
        "constant": {
            "all": list(set.union(TERMS_IN, TERMS_OUT, {"name", "entity_id", "launch_file"})),
            "actions": ["process"],
            "observations": ["process"],
        },
    },
    "term": {
        "items": {
            "repeat": ["all", "empty", "window"],
        },
        "constant": {
            "all": ["msg_type"],
            "feedthroughs": ["space_converter"],
        },
        "hide": {
            "all": ["address", "external_rate"],
            "states": ["converter"],
            "targets": ["space_converter"],
            "sensors": ["start_with_msg"],
            "actions": ["start_with_msg", "space_converter"],
            "observations": ["is_reactive", "rate", "space_converter"],
            "outputs": ["rate"],
            "inputs": ["rate"],
        },
    },
}

ENGINE_GUI_WIDGETS = deepcopy(GUI_WIDGETS)
ENGINE_GUI_WIDGETS["term"]["hide"]["all"].remove("external_rate")
ENGINE_GUI_WIDGETS["term"]["hide"]["sensors"].append("external_rate")

# Corresponding RGB values for colors
GUI_COLORS = {
    "black": [0, 0, 0],
    "red": [255, 0, 0],
    "green": [0, 128, 0],
    "yellow": [255, 255, 0],
    "blue": [0, 0, 255],
    "magenta": [255, 0, 255],
    "cyan": [0, 255, 255],
    "white": [255, 255, 255],
}

# Config files to ignore in GUI
GUI_ENTITIES_TO_IGNORE = {
    "BaseConverter",
    "SpaceConverter",
    "Converter",
    "Processor",
    "EngineNode",
    "Bridge",
    "EngineState",
}
GUI_NODE_IDS_TO_IGNORE = {
    "Observations",
    "Actions",
    "Render",
    "Supervisor",
    "Environment",
}



"""
 Adapted from https://github.com/pyqtgraph/pyqtgraph/blob/master/pyqtgraph/flowchart/Node.py
"""

import numpy as np
from functools import partial

from eagerx_gui import configuration
from eagerx_gui.utils import get_yaml_type
from eagerx_gui.gui_terminal import GuiTerminal
from eagerx_gui.pyqtgraph_utils import exception_handler, ParamWindow

from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
from pyqtgraph.graphicsItems.GraphicsObject import GraphicsObject
from pyqtgraph import functions as fn
from collections import OrderedDict

translate = QtCore.QCoreApplication.translate


class RxEngineGuiNode(QtCore.QObject):
    sigClosed = QtCore.Signal(object)
    sigRenamed = QtCore.Signal(object, object)
    sigTerminalRenamed = QtCore.Signal(object, object)  # term, oldName
    sigTerminalAdded = QtCore.Signal(object, object)  # self, term
    sigTerminalRemoved = QtCore.Signal(object, object)  # self, term

    def __init__(self, name, graph):
        QtCore.QObject.__init__(self)
        self.name = name
        self._graphics_item = None
        self.terminals = OrderedDict()
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()
        self.graph = graph
        self.exception = None

        self.allow_add_terminal = False
        self.allow_remove = False
        self.is_object = False

        self.node_type = "engine_node"
        self.__initialize_terminals()

    def __next_terminal_name(self, name):
        """Return an unused terminal name"""
        name2 = name
        i = 1
        while name2 in self.terminals:
            name2 = "%s_%d" % (name, i)
            i += 1
        return name2

    def __initialize_terminals(self):
        for terminal_type in set.union(configuration.TERMS_IN, configuration.TERMS_OUT):
            if self.node_type == "render" and terminal_type == "outputs":
                continue
            if terminal_type in self.params():
                for terminal in self.params()[terminal_type]:
                    if self.node_type in ["actions", "observations"]:
                        if terminal in self.default_params()[terminal_type]:
                            continue
                    name = terminal_type + "/" + terminal
                    self.add_terminal(name=name)
                    if self.node_type == "reset_node" and terminal_type == "outputs":
                        name = "feedthroughs/" + terminal
                        self.add_terminal(name=name)

    def params(self):
        return self._get_params(graph_backup=self.graph)

    @exception_handler
    def _get_params(self):
        assert self.name in self.graph._state["nodes"], f" No entity with name '{self.name}' in graph."
        return self.graph._state["nodes"][self.name]["config"]

    def default_params(self):
        return self.graph._state["backup"][self.name]

    def set_param(self, parameter, value):
        self._set_param(parameter, value, graph_backup=self.graph)

    def get_view(self):
        return self.graph.get_view(self.name, depth=["config"])

    @exception_handler
    def _set_param(self, parameter, value):
        self.graph.set(value, self.get_view(), parameter=parameter)
        if parameter == "color":
            self.graphics_item().set_color()

    def add_action(self):
        name = self.__next_terminal_name("outputs/action")
        self.graph._add_action(name.split("/")[-1])
        self.add_terminal(name)

    def add_observation(self):
        name = self.__next_terminal_name("inputs/observation")
        self.graph._add_observation(name.split("/")[-1])
        self.add_terminal(name)

    def add_terminal(self, name):
        """Add a new terminal to this Node with the given name.

        Causes sigTerminalAdded to be emitted."""
        name = self.__next_terminal_name(name)

        term = GuiTerminal(self, name)
        self.terminals[name] = term

        if term.is_input:
            self.inputs[name] = term
        else:
            self.outputs[name] = term

        self.graphics_item().update_terminals()
        self.sigTerminalAdded.emit(self, term)
        return term

    def graphics_item(self):
        """Return the GraphicsItem for this node."""
        if self._graphics_item is None:
            self._graphics_item = NodeGraphicsItem(self)
        return self._graphics_item

    def dependent_nodes(self):
        """Return the list of nodes which provide direct input to this node"""
        nodes = set()
        for t in self.inputs.values():
            nodes |= set([i.node for i in t.input_terminals()])
        return nodes

    def __repr__(self):
        return "<Node %s @%x>" % (self.name, id(self))

    def ctrl_widget(self):
        """Return this Node's control widget.

        By default, Nodes have no control widget. Subclasses may reimplement this
        method to provide a custom widget. This method is called by Flowcharts
        when they are constructing their Node list."""
        return None

    def connected(self, local_term, remote_term):
        """Called whenever one of this node's terminals is connected elsewhere."""
        pass

    def disconnected(self, local_term, remote_term):
        """Called whenever one of this node's terminals is disconnected from another."""
        pass

    def set_exception(self, exc):
        self.exception = exc
        self.recolor()

    def clear_exception(self):
        self.set_exception(None)

    def recolor(self):
        if self.exception is None:
            self.graphics_item().setPen(QtGui.QPen(QtGui.QColor(0, 0, 0)))
        else:
            self.graphics_item().setPen(QtGui.QPen(QtGui.QColor(150, 0, 0), 3))

    def load_state(self, state):
        pos = state["gui_state"].get("pos", (0, 0))
        self.graphics_item().setPos(*pos)

    def save_terminals(self):
        terms = OrderedDict()
        for n, t in self.terminals.items():
            terms[n] = t.save_state()
        return terms

    def clear_terminals(self):
        for t in self.terminals.values():
            t.close()
        self.terminals = OrderedDict()
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()

    def close(self):
        """Cleans up after the node--removes terminals, graphicsItem, widget"""
        self.disconnect_all()
        self.clear_terminals()
        item = self.graphics_item()
        if item.scene() is not None:
            item.scene().removeItem(item)
        self._graphics_item = None
        w = self.ctrl_widget()
        if w is not None:
            w.setParent(None)
        self.sigClosed.emit(self)

    def disconnect_all(self):
        for t in self.terminals.values():
            t.disconnect_all()


class RxGuiNode(QtCore.QObject):
    sigClosed = QtCore.Signal(object)
    sigRenamed = QtCore.Signal(object, object)
    sigTerminalRenamed = QtCore.Signal(object, object)  # term, oldName
    sigTerminalAdded = QtCore.Signal(object, object)  # self, term
    sigTerminalRemoved = QtCore.Signal(object, object)  # self, term

    def __init__(self, name, graph):
        QtCore.QObject.__init__(self)
        self.name = name
        self._graphics_item = None
        self.terminals = OrderedDict()
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()
        self.graph = graph
        self.exception = None

        if self.params()["entity_id"] in ["Actions", "Observations", "Render"]:
            self.node_type = self.params()["entity_id"].lower()
            self.allow_add_terminal = not self.node_type == "render"
            self.allow_remove = False
            self.is_object = False
        else:
            self.node_type = get_yaml_type(self.default_params())
            self.allow_add_terminal = self.is_object = self.node_type == "object"
            self.allow_remove = True
        self.__initialize_terminals()

    def __next_terminal_name(self, name):
        """Return an unused terminal name"""
        name2 = name
        i = 1
        while name2 in self.terminals:
            name2 = "%s_%d" % (name, i)
            i += 1
        return name2

    def __initialize_terminals(self):
        for terminal_type in set.union(configuration.TERMS_IN, configuration.TERMS_OUT):
            if self.node_type == "render" and terminal_type == "outputs":
                continue
            if terminal_type in self.params():
                for terminal in self.params()[terminal_type]:
                    if self.node_type in ["actions", "observations"]:
                        if terminal in self.default_params()[terminal_type]:
                            continue
                    name = terminal_type + "/" + terminal
                    self.add_terminal(name=name)
                    if self.node_type == "reset_node" and terminal_type == "outputs":
                        name = "feedthroughs/" + terminal
                        self.add_terminal(name=name)

    def params(self):
        return self._get_params(graph_backup=self.graph)

    @exception_handler
    def _get_params(self):
        assert self.name in self.graph._state["nodes"], f" No entity with name '{self.name}' in graph."
        return self.graph._state["nodes"][self.name]["config"]

    def default_params(self):
        return self.graph._state["backup"][self.name]

    def set_param(self, parameter, value):
        self._set_param(parameter, value, graph_backup=self.graph)

    def get_view(self):
        return self.graph.get_view(self.name, depth=["config"])

    @exception_handler
    def _set_param(self, parameter, value):
        self.graph.set(value, self.get_view(), parameter=parameter)
        if parameter == "color":
            self.graphics_item().set_color()

    def remove_terminal(self, term):
        """Remove the specified terminal from this Node. May specify either the
        terminal's name or the terminal itself.

        Causes sigTerminalRemoved to be emitted."""
        if isinstance(term, GuiTerminal):
            name = term.name
        else:
            name = term
            term = self.terminals[name]
        term.close()

        del self.terminals[name]
        if term.is_input:
            del self.inputs[name]
        else:
            del self.outputs[name]

        self.graphics_item().update_terminals()
        self.sigTerminalRemoved.emit(self, term)

    def terminal_renamed(self, term, old_name):
        """Called after a terminal has been renamed

        Causes sigTerminalRenamed to be emitted."""
        new_name = term.name
        for d in [self.terminals, self.inputs, self.outputs]:
            if old_name not in d:
                continue
            d[new_name] = d[old_name]
            del d[old_name]
        gui_state = self.graph._state["gui_state"][self.name]
        if old_name in gui_state["linestyle"]:
            gui_state["linestyle"][new_name] = gui_state["linestyle"][old_name]
            gui_state["linestyle"].pop(old_name)
        self.graphics_item().update_terminals()
        self.sigTerminalRenamed.emit(term, old_name)

    def add_action(self):
        name = self.__next_terminal_name("outputs/action")
        self.graph._add_action(name.split("/")[-1])
        self.add_terminal(name)

    def add_observation(self):
        name = self.__next_terminal_name("inputs/observation")
        self.graph._add_observation(name.split("/")[-1])
        self.add_terminal(name)

    def add_terminal(self, name):
        """Add a new terminal to this Node with the given name.

        Causes sigTerminalAdded to be emitted."""
        name = self.__next_terminal_name(name)

        term = GuiTerminal(self, name)
        self.terminals[name] = term

        if term.is_input:
            self.inputs[name] = term
        else:
            self.outputs[name] = term

        self.graphics_item().update_terminals()
        self.sigTerminalAdded.emit(self, term)
        return term

    def graphics_item(self):
        """Return the GraphicsItem for this node."""
        if self._graphics_item is None:
            self._graphics_item = NodeGraphicsItem(self)
        return self._graphics_item

    def rename(self, name):
        """Rename this node. This will cause sigRenamed to be emitted."""
        self.graph.rename(self.name, name)
        old_name = self.name
        self.name = name
        self.sigRenamed.emit(self, old_name)

    def dependent_nodes(self):
        """Return the list of nodes which provide direct input to this node"""
        nodes = set()
        for t in self.inputs.values():
            nodes |= set([i.node for i in t.input_terminals()])
        return nodes

    def __repr__(self):
        return "<Node %s @%x>" % (self.name, id(self))

    def ctrl_widget(self):
        """Return this Node's control widget.

        By default, Nodes have no control widget. Subclasses may reimplement this
        method to provide a custom widget. This method is called by Flowcharts
        when they are constructing their Node list."""
        return None

    def connected(self, local_term, remote_term):
        """Called whenever one of this node's terminals is connected elsewhere."""
        pass

    def disconnected(self, local_term, remote_term):
        """Called whenever one of this node's terminals is disconnected from another."""
        pass

    def set_exception(self, exc):
        self.exception = exc
        self.recolor()

    def clear_exception(self):
        self.set_exception(None)

    def recolor(self):
        if self.exception is None:
            self.graphics_item().setPen(QtGui.QPen(QtGui.QColor(0, 0, 0)))
        else:
            self.graphics_item().setPen(QtGui.QPen(QtGui.QColor(150, 0, 0), 3))

    def load_state(self, state):
        pos = state["gui_state"].get("pos", (0, 0))
        self.graphics_item().setPos(*pos)

    def save_terminals(self):
        terms = OrderedDict()
        for n, t in self.terminals.items():
            terms[n] = t.save_state()
        return terms

    def clear_terminals(self):
        for t in self.terminals.values():
            t.close()
        self.terminals = OrderedDict()
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()

    def close(self):
        """Cleans up after the node--removes terminals, graphicsItem, widget"""
        self.disconnect_all()
        self.clear_terminals()
        item = self.graphics_item()
        if item.scene() is not None:
            item.scene().removeItem(item)
        self._graphics_item = None
        w = self.ctrl_widget()
        if w is not None:
            w.setParent(None)
        self.sigClosed.emit(self)

    def disconnect_all(self):
        for t in self.terminals.values():
            t.disconnect_all()


class TextItem(QtWidgets.QGraphicsTextItem):
    def __init__(self, text, parent, on_update):
        super().__init__(text, parent)
        self.on_update = on_update

    def focusOutEvent(self, ev):
        super().focusOutEvent(ev)
        if self.on_update is not None:
            self.on_update()

    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            if self.on_update is not None:
                self.on_update()
                return
        super().keyPressEvent(ev)

    def mousePressEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            self.setFocus(QtCore.Qt.FocusReason.MouseFocusReason)  # focus text label


class NodeGraphicsItem(GraphicsObject):
    def __init__(self, node):
        GraphicsObject.__init__(self)
        self.node = node
        self._node_type = node.node_type
        flags = (
            self.GraphicsItemFlag.ItemIsMovable
            | self.GraphicsItemFlag.ItemIsSelectable
            | self.GraphicsItemFlag.ItemIsFocusable
            | self.GraphicsItemFlag.ItemSendsGeometryChanges
        )

        self.set_color()
        self.setFlags(flags)
        self.bounds = QtCore.QRectF(0, 0, 125, 125)
        self.nameItem = TextItem(self.node.name, self, self.label_changed)
        self.nameItem.setDefaultTextColor(QtGui.QColor(50, 50, 50))
        self.nameItem.moveBy(self.bounds.width() / 2.0 - self.nameItem.boundingRect().width() / 2.0, 0)
        self.update_terminals()

        self.nameItem.focusOutEvent = self.label_focus_out
        self.nameItem.keyPressEvent = self.label_key_press

        self.menu = None
        self.buildMenu()
        self.initial_z_value = self.zValue()
        self.label_changed()

    def set_color(self):
        if "color" in self.node.params() and self.node.params()["color"] in configuration.GUI_COLORS:
            brush_color = np.array(configuration.GUI_COLORS[self.node.params()["color"]])
        else:
            brush_color = np.array([200, 200, 200])

        self.brush = fn.mkBrush(*brush_color, 50)
        self.hoverBrush = fn.mkBrush(*brush_color, 100)
        self.selectBrush = fn.mkBrush(*brush_color, 100)
        self.pen = fn.mkPen(0, 0, 0, 200, width=2)
        self.selectPen = fn.mkPen(0, 0, 0, 200, width=4)
        self.hovered = False
        self.update()

    def label_focus_out(self, ev):
        QtWidgets.QGraphicsTextItem.focusOutEvent(self.nameItem, ev)
        self.label_changed()

    def label_key_press(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Enter or ev.key() == QtCore.Qt.Key.Key_Return:
            self.label_changed()
        else:
            QtWidgets.QGraphicsTextItem.keyPressEvent(self.nameItem, ev)

    def label_changed(self):
        new_name = str(self.nameItem.toPlainText())
        if self.node.graph is not None and new_name in self.node.graph.nodes.keys():
            self.nameItem.setPlainText(self.node.name)
            return
        if new_name != self.node.name:
            self.node.rename(new_name)
        self.update_terminals()

    def setPen(self, *args, **kwargs):
        self.pen = fn.mkPen(*args, **kwargs)
        self.update()

    def setBrush(self, brush):
        self.brush = brush
        self.update()

    def update_terminals(self):
        # re-center the label
        if self.nameItem.boundingRect().width() > self.bounds.width() - 5.0:
            self.bounds = QtCore.QRectF(0, 0, self.nameItem.boundingRect().width() + 5.0, self.bounds.height())
        bounds = self.boundingRect()
        self.nameItem.setPos(bounds.width() / 2.0 - self.nameItem.boundingRect().width() / 2.0, 0)

        # Update bounds
        bounds = self.bounds
        self.terminals = {}
        inp = self.node.inputs
        y = self.nameItem.boundingRect().height() + 5
        max_width_inp = 0.0
        for i, t in inp.items():
            item = t.graphics_item()
            item.setParentItem(self)
            br = self.bounds
            item.set_anchor(0, y)
            self.terminals[i] = (t, item)
            y += item.boundingRect().height() + 2
            if item.boundingRect().width() > max_width_inp:
                max_width_inp = item.boundingRect().width()
        y_inp = y

        out = self.node.outputs
        dy = bounds.height() / (len(out) + 1)
        y = self.nameItem.boundingRect().height() + 5
        max_width_out = 0.0
        for i, t in out.items():
            item = t.graphics_item()
            item.setParentItem(self)
            item.setZValue(self.initial_z_value)
            br = self.bounds
            item.set_anchor(bounds.width(), y)
            self.terminals[i] = (t, item)
            y += item.boundingRect().height() + 2
            if item.boundingRect().width() > max_width_out:
                max_width_out = item.boundingRect().width()
        y_max = max(y_inp, y)

        if y_max > bounds.height() - 5:
            self.bounds = QtCore.QRectF(0, 0, self.bounds.width(), y_max + 6.0)

        if max_width_inp + max_width_out > bounds.width() + 10:
            self.bounds = QtCore.QRectF(0, 0, max_width_inp + max_width_out + 10.0, self.bounds.height())
            self.update_terminals()
        self.update()

    def boundingRect(self):
        return self.bounds.adjusted(-5, -5, 5, 5)

    def paint(self, p, *args):
        p.setPen(self.pen)
        if self.isSelected():
            p.setPen(self.selectPen)
            p.setBrush(self.selectBrush)
            self.setZValue(200)
        else:
            p.setPen(self.pen)
            if self.hovered:
                p.setBrush(self.hoverBrush)
                self.setZValue(200)
            else:
                p.setBrush(self.brush)
                self.setZValue(self.initial_z_value)

        p.drawRect(self.bounds)

    def mousePressEvent(self, ev):
        ev.ignore()

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            sel = self.isSelected()
            self.setSelected(True)
            if not sel and self.isSelected():
                self.update()

        elif ev.button() == QtCore.Qt.MouseButton.RightButton:
            self.menu = None
            self.buildMenu()
            ev.accept()
            self.raiseContextMenu(ev)

    def mouseDoubleClickEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            param_window = ParamWindow(node=self.node)
            param_window.open()
            param_window.close()

    def mouseDragEvent(self, ev):
        if ev.button() == QtCore.Qt.MouseButton.LeftButton:
            ev.accept()
            self.setPos(self.pos() + self.mapToParent(ev.pos()) - self.mapToParent(ev.lastPos()))

    def hoverEvent(self, ev):
        if not ev.isExit() and ev.acceptClicks(QtCore.Qt.MouseButton.LeftButton):
            ev.acceptDrags(QtCore.Qt.MouseButton.LeftButton)
            self.hovered = True
        else:
            self.hovered = False
        self.update()

    def keyPressEvent(self, ev):
        if ev.key() == QtCore.Qt.Key.Key_Delete or ev.key() == QtCore.Qt.Key.Key_Backspace:
            ev.accept()
            if not self.node.allow_remove:
                return
            self.node.graph.remove(self.node.name, remove=False)
            if self.node.name in self.node.graph._state["gui_state"]:
                # Pop node from gui_state upon removal
                self.node.graph._state["gui_state"].pop(self.node.name)
            self.node.close()
        else:
            ev.ignore()

    def itemChange(self, change, val):
        if change == self.GraphicsItemChange.ItemPositionHasChanged:
            self.node.graph._state["gui_state"][self.node.name]["pos"] = [
                self.pos().x(),
                self.pos().y(),
            ]
            for k, t in self.terminals.items():
                t[1].node_moved()
        return GraphicsObject.itemChange(self, change, val)

    def getMenu(self):
        return self.menu

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))

    def buildMenu(self):
        self.menu = QtWidgets.QMenu()
        self.menu.setTitle("Node")

        if self.node.allow_add_terminal:
            if self._node_type == "observations":
                self.menu.addAction("Add observation", self.node.add_observation)
            elif self._node_type == "actions":
                self.menu.addAction("Add action", self.node.add_action)
            else:
                for terminal_type in set.union(
                    configuration.TERMS[self._node_type]["in"],
                    configuration.TERMS[self._node_type]["out"],
                ):
                    if terminal_type in self.node.default_params():
                        terminal_menu = QtWidgets.QMenu("Add {}".format(terminal_type[:-1]), self.menu)
                        for terminal in self.node.default_params()[terminal_type]:
                            terminal_name = terminal_type + "/" + terminal
                            act = terminal_menu.addAction(
                                terminal,
                                partial(
                                    self.add_terminal,
                                    terminal_type=terminal_type,
                                    terminal_name=terminal,
                                ),
                            )
                            if terminal_name in self.node.terminals.keys():
                                act.setEnabled(False)
                            self.menu.addMenu(terminal_menu)
        if self.node.allow_remove:
            self.menu.addAction("Remove {}".format(self.node.name), self.remove_node)

    def add_terminal(self, terminal_type, terminal_name):
        view = self.node.graph.get_view(self.node.name, depth=[terminal_type, terminal_name])
        self.node.graph._add_component(view)
        name = terminal_type + "/" + terminal_name
        self.node.add_terminal(name=name)

    def remove_node(self):
        self.node.graph.remove(self.node.name, remove=False)
        self.node.close()



__version__ = "0.1.6"

import os

os.environ["PYQTGRAPH_QT_LIB"] = "PyQt6"

import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets
from eagerx_gui.gui import Gui, EngineGui
import sys


def launch_gui(state):
    app = QtWidgets.QApplication(sys.argv)

    ## Create main window with grid layout
    win = QtWidgets.QMainWindow()
    win.setWindowTitle("EAGERx Graph")
    cw = QtWidgets.QWidget()
    win.setCentralWidget(cw)
    layout = QtWidgets.QGridLayout()
    cw.setLayout(layout)

    rx_gui = Gui(state)
    w = rx_gui.widget()

    # Add flowchart control panel to the main window
    layout.addWidget(w, 0, 0, 2, 1)

    win.show()

    app.exec()
    new_state = rx_gui.state()
    app.quit()
    return new_state


def launch_engine_gui(state):
    app = QtWidgets.QApplication(sys.argv)

    ## Create main window with grid layout
    win = QtWidgets.QMainWindow()
    win.setWindowTitle("EAGERx EngineGraph")
    cw = QtWidgets.QWidget()
    win.setCentralWidget(cw)
    layout = QtWidgets.QGridLayout()
    cw.setLayout(layout)

    rx_gui = EngineGui(state)
    w = rx_gui.widget()

    # Add flowchart control panel to the main window
    layout.addWidget(w, 0, 0, 2, 1)

    win.show()

    app.exec()
    new_state = rx_gui.state()
    app.quit()
    return new_state
